/***************************************************************************
                          Ellpack.h  -  description
                             -------------------
    begin                : Dec 7, 2013
    copyright            : (C) 2013 by Tomas Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <TNL/Matrices/Sparse.h>
#include <TNL/Containers/Vector.h>

namespace TNL {
namespace Matrices {   

template< typename Execution >
class EllpackExecutionDependentCode;

template< typename Real, typename Execution = TNL::Execution::OpenMP, typename Index = int >
class Ellpack : public Sparse< Real, Execution, Index >
{
private:
   // convenient template alias for controlling the selection of copy-assignment operator
   template< typename Execution2 >
   using Enabler = std::enable_if< ! std::is_same< Execution2, Execution >::value >;

   // friend class will be needed for templated assignment operators
   template< typename Real2, typename Execution2, typename Index2 >
   friend class Ellpack;

public:
   typedef Real RealType;
   typedef Execution ExecutionType;
   typedef Index IndexType;
   typedef typename Sparse< RealType, ExecutionType, IndexType >::CompressedRowLengthsVector CompressedRowLengthsVector;
   typedef typename Sparse< RealType, ExecutionType, IndexType >::ConstCompressedRowLengthsVectorView ConstCompressedRowLengthsVectorView;
   typedef typename Sparse< RealType, ExecutionType, IndexType >::ValuesVector ValuesVector;
   typedef typename Sparse< RealType, ExecutionType, IndexType >::ColumnIndexesVector ColumnIndexesVector;
   typedef Sparse< Real, Execution, Index > BaseType;
   typedef typename BaseType::MatrixRow MatrixRow;
   typedef SparseRow< const RealType, const IndexType > ConstMatrixRow;

   template< typename _Value,
             typename _Execution = Execution,
             typename _Index = Index >
   using Self = Ellpack< _Value, _Execution, _Index >;

   Ellpack();

   static String getSerializationType();

   virtual String getSerializationTypeVirtual() const;

   void setDimensions( const IndexType rows,
                       const IndexType columns );

   void setCompressedRowLengths( ConstCompressedRowLengthsVectorView rowLengths );

   void setConstantCompressedRowLengths( const IndexType& rowLengths );

   IndexType getRowLength( const IndexType row ) const;

   __cuda_callable__
   IndexType getRowLengthFast( const IndexType row ) const;
   
   IndexType getNonZeroRowLength( const IndexType row ) const;

   template< typename Real2, typename Execution2, typename Index2 >
   void setLike( const Ellpack< Real2, Execution2, Index2 >& matrix );

   void reset();
 
   template< typename Real2, typename Execution2, typename Index2 >
   bool operator == ( const Ellpack< Real2, Execution2, Index2 >& matrix ) const;

   template< typename Real2, typename Execution2, typename Index2 >
   bool operator != ( const Ellpack< Real2, Execution2, Index2 >& matrix ) const;

   __cuda_callable__
   bool setElementFast( const IndexType row,
                        const IndexType column,
                        const RealType& value );

   bool setElement( const IndexType row,
                    const IndexType column,
                    const RealType& value );

   __cuda_callable__
   bool addElementFast( const IndexType row,
                        const IndexType column,
                        const RealType& value,
                        const RealType& thisElementMultiplicator = 1.0 );

   bool addElement( const IndexType row,
                    const IndexType column,
                    const RealType& value,
                    const RealType& thisElementMultiplicator = 1.0 );


   __cuda_callable__
   bool setRowFast( const IndexType row,
                    const IndexType* columnIndexes,
                    const RealType* values,
                    const IndexType elements );

   bool setRow( const IndexType row,
                const IndexType* columnIndexes,
                const RealType* values,
                const IndexType elements );


   __cuda_callable__
   bool addRowFast( const IndexType row,
                    const IndexType* columns,
                    const RealType* values,
                    const IndexType numberOfElements,
                    const RealType& thisElementMultiplicator = 1.0 );

   bool addRow( const IndexType row,
                const IndexType* columns,
                const RealType* values,
                const IndexType numberOfElements,
                const RealType& thisElementMultiplicator = 1.0 );

   __cuda_callable__
   RealType getElementFast( const IndexType row,
                            const IndexType column ) const;

   RealType getElement( const IndexType row,
                        const IndexType column ) const;

   __cuda_callable__
   void getRowFast( const IndexType row,
                    IndexType* columns,
                    RealType* values ) const;

   __cuda_callable__
   MatrixRow getRow( const IndexType rowIndex );

   __cuda_callable__
   ConstMatrixRow getRow( const IndexType rowIndex ) const;

   template< typename Vector >
   __cuda_callable__
   typename Vector::RealType rowVectorProduct( const IndexType row,
                                               const Vector& vector ) const;

   template< typename InVector,
             typename OutVector >
   void vectorProduct( const InVector& inVector,
                       OutVector& outVector,
                       RealType multiplicator = 1.0 ) const;

   template< typename Real2, typename Index2 >
   void addMatrix( const Ellpack< Real2, Execution, Index2 >& matrix,
                   const RealType& matrixMultiplicator = 1.0,
                   const RealType& thisMatrixMultiplicator = 1.0 );

   template< typename Real2, typename Index2 >
   void getTransposition( const Ellpack< Real2, Execution, Index2 >& matrix,
                          const RealType& matrixMultiplicator = 1.0 );

   template< typename Vector1, typename Vector2 >
   bool performSORIteration( const Vector1& b,
                             const IndexType row,
                             Vector2& x,
                             const RealType& omega = 1.0 ) const;

   template< typename Vector >
   bool performJacobiIteration( const Vector& b,
								const IndexType row,
								const Vector& old_x,
								Vector& x,
								const RealType& omega ) const;
   
   // copy assignment
   Ellpack& operator=( const Ellpack& matrix );   

   // cross-device copy assignment
   template< typename Real2, typename Execution2, typename Index2,
             typename = typename Enabler< Execution2 >::type >
   Ellpack& operator=( const Ellpack< Real2, Execution2, Index2 >& matrix );

   void save( File& file ) const;

   void load( File& file );

   void save( const String& fileName ) const;

   void load( const String& fileName );

   void print( std::ostream& str ) const;

protected:

   void allocateElements();

   IndexType rowLengths, alignedRows;

   typedef EllpackExecutionDependentCode< ExecutionType > ExecutionDependentCode;
   friend class EllpackExecutionDependentCode< ExecutionType >;
};

} // namespace Matrices
} // namespace TNL

#include <TNL/Matrices/Ellpack_impl.h>
