/***************************************************************************
                          MeshEntity_impl.h  -  description
                             -------------------
    begin                : Sep 8, 2015
    copyright            : (C) 2015 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

/***
 * Authors:
 * Oberhuber Tomas, tomas.oberhuber@fjfi.cvut.cz
 * Zabka Vitezslav, zabkav@gmail.com
 */

#pragma once

#include <TNL/Meshes/MeshEntity.h>

namespace TNL {
namespace Meshes {

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
__cuda_callable__
MeshEntity< MeshConfig, Execution, EntityTopology >::
MeshEntity( const MeshEntity& entity )
   : SubentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >( entity ),
     SuperentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >( entity ),
     MeshEntityIndex< typename MeshConfig::IdType >( entity )
{
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
   template< typename Execution_ >
MeshEntity< MeshConfig, Execution, EntityTopology >::
MeshEntity( const MeshEntity< MeshConfig, Execution_, EntityTopology >& entity )
   // no cross-device copy of subentities and superentities here - Mesh constructor has to rebind pointers
   : MeshEntityIndex< typename MeshConfig::IdType >( entity )
{
   static_assert( ! std::is_same< Execution, Execution_ >::value, "this should never happen" );
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
__cuda_callable__
MeshEntity< MeshConfig, Execution, EntityTopology >& 
MeshEntity< MeshConfig, Execution, EntityTopology >::
operator=( const MeshEntity& entity )
{
   SubentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >::operator=( entity );
   SuperentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >::operator=( entity );
   MeshEntityIndex< typename MeshConfig::IdType >::operator=( entity );
   return *this;
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
   template< typename Execution_ >
__cuda_callable__
MeshEntity< MeshConfig, Execution, EntityTopology >& 
MeshEntity< MeshConfig, Execution, EntityTopology >::
operator=( const MeshEntity< MeshConfig, Execution_, EntityTopology >& entity )
{
   static_assert( ! std::is_same< Execution, Execution_ >::value, "this should never happen" );

   // no cross-device copy of subentities and superentities here - Mesh::operator= has to rebind pointers
   MeshEntityIndex< typename MeshConfig::IdType >::operator=( entity );
   return *this;
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
String
MeshEntity< MeshConfig, Execution, EntityTopology >::
getSerializationType()
{
   return String( "MeshEntity<" ) +
          TNL::getSerializationType< MeshConfig >() + ", " +
          TNL::getSerializationType< EntityTopology >() + ">";
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
String
MeshEntity< MeshConfig, Execution, EntityTopology >::
getSerializationTypeVirtual() const
{
   return this->getSerializationType();
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
void
MeshEntity< MeshConfig, Execution, EntityTopology >::
save( File& file ) const
{
   // no I/O for subentities and superentities - not loaded anyway
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
void
MeshEntity< MeshConfig, Execution, EntityTopology >::
load( File& file )
{
   // no I/O for subentities and superentities - Mesh::load has to rebind pointers
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
void
MeshEntity< MeshConfig, Execution, EntityTopology >::
print( std::ostream& str ) const
{
   str << "\t Mesh entity dimension: " << EntityTopology::dimension << std::endl;
   SubentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >::print( str );
   SuperentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >::print( str );
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
__cuda_callable__
bool
MeshEntity< MeshConfig, Execution, EntityTopology >::
operator==( const MeshEntity& entity ) const
{
   return ( SubentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >::operator==( entity ) &&
            SuperentityAccessLayerFamily< MeshConfig, Execution, EntityTopology >::operator==( entity ) &&
            MeshEntityIndex< typename MeshConfig::IdType >::operator==( entity ) );
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
__cuda_callable__
bool
MeshEntity< MeshConfig, Execution, EntityTopology >::
operator!=( const MeshEntity& entity ) const
{
   return ! ( *this == entity );
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
constexpr int
MeshEntity< MeshConfig, Execution, EntityTopology >::
getEntityDimension()
{
   return EntityTopology::dimension;
}

/****
 * Subentities
 */
template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
constexpr typename MeshEntity< MeshConfig, Execution, EntityTopology >::LocalIndexType
MeshEntity< MeshConfig, Execution, EntityTopology >::
getVerticesCount()
{
   return SubentityTraits< 0 >::count;
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
typename MeshEntity< MeshConfig, Execution, EntityTopology >::GlobalIndexType
MeshEntity< MeshConfig, Execution, EntityTopology >::
getVertexIndex( const LocalIndexType localIndex ) const
{
   return this->template getSubentityIndex< 0 >( localIndex  );
}


/****
 * Vertex entity specialization
 */
template< typename MeshConfig, typename Execution >
__cuda_callable__
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
MeshEntity( const MeshEntity& entity )
   : SuperentityAccessLayerFamily< MeshConfig, Execution, Topologies::Vertex >( entity ),
     MeshEntityIndex< typename MeshConfig::IdType >( entity )
{
   setPoint( entity.getPoint() );
}

template< typename MeshConfig, typename Execution >
   template< typename Execution_ >
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
MeshEntity( const MeshEntity< MeshConfig, Execution_, Topologies::Vertex >& entity )
   // no cross-device copy of superentities here - Mesh constructor has to rebind pointers
   : MeshEntityIndex< typename MeshConfig::IdType >( entity )
{
   static_assert( ! std::is_same< Execution, Execution_ >::value, "this should never happen" );

   setPoint( entity.getPoint() );
}

template< typename MeshConfig, typename Execution >
__cuda_callable__
MeshEntity< MeshConfig, Execution, Topologies::Vertex >& 
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
operator=( const MeshEntity& entity )
{
   SuperentityAccessLayerFamily< MeshConfig, Execution, Topologies::Vertex >::operator=( entity );
   MeshEntityIndex< typename MeshConfig::IdType >::operator=( entity );
   setPoint( entity.getPoint() );
   return *this;
}

template< typename MeshConfig, typename Execution >
   template< typename Execution_ >
__cuda_callable__
MeshEntity< MeshConfig, Execution, Topologies::Vertex >& 
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
operator=( const MeshEntity< MeshConfig, Execution_, Topologies::Vertex >& entity )
{
   static_assert( ! std::is_same< Execution, Execution_ >::value, "this should never happen" );

   // no cross-device copy of superentities here - Mesh::operator= has to rebind pointers
   MeshEntityIndex< typename MeshConfig::IdType >::operator=( entity );
   setPoint( entity.getPoint() );
   return *this;
}

template< typename MeshConfig, typename Execution >
String
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
getSerializationType()
{
   return String( "MeshEntity<" ) +
          TNL::getSerializationType< MeshConfig >() + ", " +
          TNL::getSerializationType< Topologies::Vertex >() + ">";
}

template< typename MeshConfig, typename Execution >
String
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
getSerializationTypeVirtual() const
{
   return this->getSerializationType();
}

template< typename MeshConfig, typename Execution >
void
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
save( File& file ) const
{
   // no I/O for superentities - not loaded anyway
   point.save( file );
}

template< typename MeshConfig, typename Execution >
void
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
load( File& file )
{
   // no I/O for superentities - Mesh::load has to rebind pointers
   point.load( file );
}

template< typename MeshConfig, typename Execution >
void
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
print( std::ostream& str ) const
{
   str << "\t Mesh entity dimension: " << Topologies::Vertex::dimension << std::endl;
   str << "\t Coordinates = " << point << std::endl;
   SuperentityAccessLayerFamily< MeshConfig, Execution, Topologies::Vertex >::print( str );
}

template< typename MeshConfig, typename Execution >
__cuda_callable__
bool
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
operator==( const MeshEntity& entity ) const
{
   return ( SuperentityAccessLayerFamily< MeshConfig, Execution, Topologies::Vertex >::operator==( entity ) &&
            MeshEntityIndex< typename MeshConfig::IdType >::operator==( entity ) &&
            point == entity.point );
}

template< typename MeshConfig, typename Execution >
__cuda_callable__
bool
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
operator!=( const MeshEntity& entity ) const
{
   return ! ( *this == entity );
}

template< typename MeshConfig, typename Execution >
constexpr int
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
getEntityDimension()
{
   return EntityTopology::dimension;
}

template< typename MeshConfig, typename Execution >
__cuda_callable__
typename MeshEntity< MeshConfig, Execution, Topologies::Vertex >::PointType
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
getPoint() const
{
   return this->point;
}

template< typename MeshConfig, typename Execution >
__cuda_callable__
void
MeshEntity< MeshConfig, Execution, Topologies::Vertex >::
setPoint( const PointType& point )
{
   this->point = point;
}

template< typename MeshConfig,
          typename Execution,
          typename EntityTopology >
std::ostream& operator<<( std::ostream& str, const MeshEntity< MeshConfig, Execution, EntityTopology >& entity )
{
   entity.print( str );
   return str;
}

} // namespace Meshes
} // namespace TNL
