/***************************************************************************
                          LayerFamily.h  -  description
                             -------------------
    begin                : Nov 9, 2017
    copyright            : (C) 2017 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include "ConfigValidator.h"
#include "Initializer.h"
#include "Layer.h"
#include "Traits.h"

namespace TNL {
namespace Meshes {
namespace BoundaryTags {

template< typename MeshConfig, typename Execution, typename Dimension = DimensionTag< 0 > >
class LayerInheritor
   : public Layer< MeshConfig, Execution, Dimension >,
     public LayerInheritor< MeshConfig, Execution, typename Dimension::Increment >
{
   using LayerType = Layer< MeshConfig, Execution, Dimension >;
   using BaseType = LayerInheritor< MeshConfig, Execution, typename Dimension::Increment >;
protected:
   using LayerType::setEntitiesCount;
   using LayerType::resetBoundaryTags;
   using LayerType::setIsBoundaryEntity;
   using LayerType::isBoundaryEntity;
   using LayerType::updateBoundaryIndices;
   using LayerType::getBoundaryEntitiesCount;
   using LayerType::getBoundaryEntityIndex;
   using LayerType::getInteriorEntitiesCount;
   using LayerType::getInteriorEntityIndex;

   using BaseType::setEntitiesCount;
   using BaseType::resetBoundaryTags;
   using BaseType::setIsBoundaryEntity;
   using BaseType::isBoundaryEntity;
   using BaseType::updateBoundaryIndices;
   using BaseType::getBoundaryEntitiesCount;
   using BaseType::getBoundaryEntityIndex;
   using BaseType::getInteriorEntitiesCount;
   using BaseType::getInteriorEntityIndex;


   LayerInheritor() = default;

   explicit LayerInheritor( const LayerInheritor& other )
   {
      operator=( other );
   }

   template< typename Execution_ >
   LayerInheritor( const LayerInheritor< MeshConfig, Execution_, Dimension >& other )
   {
      operator=( other );
   }

   LayerInheritor& operator=( const LayerInheritor& other )
   {
      LayerType::operator=( other );
      BaseType::operator=( other );
      return *this;
   }

   template< typename Execution_ >
   LayerInheritor& operator=( const LayerInheritor< MeshConfig, Execution_, Dimension >& other )
   {
      LayerType::operator=( other );
      BaseType::operator=( other );
      return *this;
   }


   void save( File& file ) const
   {
      LayerType::save( file );
      BaseType::save( file );
   }

   void load( File& file )
   {
      LayerType::load( file );
      BaseType::load( file );
   }

   void print( std::ostream& str ) const
   {
      LayerType::print( str );
      BaseType::print( str );
   }

   bool operator==( const LayerInheritor& layer ) const
   {
      return LayerType::operator==( layer ) &&
             BaseType::operator==( layer );
   }
};

template< typename MeshConfig, typename Execution >
class LayerInheritor< MeshConfig, Execution, DimensionTag< MeshConfig::meshDimension + 1 > >
{
protected:
   void setEntitiesCount();
   void resetBoundaryTags();
   void setIsBoundaryEntity();
   void isBoundaryEntity();
   void updateBoundaryIndices();
   void getBoundaryEntitiesCount();
   void getBoundaryEntityIndex();
   void getInteriorEntitiesCount();
   void getInteriorEntityIndex();

   LayerInheritor() = default;
   explicit LayerInheritor( const LayerInheritor& other ) {}
   template< typename Execution_ >
   LayerInheritor( const LayerInheritor< MeshConfig, Execution_, DimensionTag< MeshConfig::meshDimension + 1 > >& other ) {}
   LayerInheritor& operator=( const LayerInheritor& other ) { return *this; }
   template< typename Execution_ >
   LayerInheritor& operator=( const LayerInheritor< MeshConfig, Execution_, DimensionTag< MeshConfig::meshDimension + 1 > >& other ) { return *this; }

   void save( File& file ) const {}
   void load( File& file ) {}

   void print( std::ostream& str ) const {}

   bool operator==( const LayerInheritor& layer ) const
   {
      return true;
   }
};

// Note that MeshType is an incomplete type and therefore cannot be used to access
// MeshType::Config etc. at the time of declaration of this class template.
template< typename MeshConfig, typename Execution, typename MeshType >
class LayerFamily
   : public ConfigValidator< MeshConfig >,
     public Initializer< MeshConfig, Execution, MeshType >,
     public LayerInheritor< MeshConfig, Execution >
{
   using MeshTraitsType = MeshTraits< MeshConfig, Execution >;
   using GlobalIndexType = typename MeshTraitsType::GlobalIndexType;
   using BaseType = LayerInheritor< MeshConfig, Execution, DimensionTag< 0 > >;
   template< int Dimension >
   using EntityTraits = typename MeshTraitsType::template EntityTraits< Dimension >;
   template< int Dimension >
   using WeakTrait = WeakStorageTrait< MeshConfig, Execution, DimensionTag< Dimension > >;

   friend Initializer< MeshConfig, Execution, MeshType >;

public:
   // inherit constructors and assignment operators (including templated versions)
   using BaseType::BaseType;
   using BaseType::operator=;

   // getters for boundary tags
   template< int Dimension >
   __cuda_callable__
   bool isBoundaryEntity( const GlobalIndexType& entityIndex ) const
   {
      static_assert( WeakTrait< Dimension >::boundaryTagsEnabled, "You try to access boundary tags which are not configured for storage." );
      return BaseType::isBoundaryEntity( DimensionTag< Dimension >(), entityIndex );
   }

   template< int Dimension >
   __cuda_callable__
   GlobalIndexType getBoundaryEntitiesCount() const
   {
      static_assert( WeakTrait< Dimension >::boundaryTagsEnabled, "You try to access boundary tags which are not configured for storage." );
      return BaseType::getBoundaryEntitiesCount( DimensionTag< Dimension >() );
   }

   template< int Dimension >
   __cuda_callable__
   GlobalIndexType getBoundaryEntityIndex( const GlobalIndexType& i ) const
   {
      static_assert( WeakTrait< Dimension >::boundaryTagsEnabled, "You try to access boundary tags which are not configured for storage." );
      return BaseType::getBoundaryEntityIndex( DimensionTag< Dimension >(), i );
   }

   template< int Dimension >
   __cuda_callable__
   GlobalIndexType getInteriorEntitiesCount() const
   {
      static_assert( WeakTrait< Dimension >::boundaryTagsEnabled, "You try to access boundary tags which are not configured for storage." );
      return BaseType::getInteriorEntitiesCount( DimensionTag< Dimension >() );
   }

   template< int Dimension >
   __cuda_callable__
   GlobalIndexType getInteriorEntityIndex( const GlobalIndexType& i ) const
   {
      static_assert( WeakTrait< Dimension >::boundaryTagsEnabled, "You try to access boundary tags which are not configured for storage." );
      return BaseType::getInteriorEntityIndex( DimensionTag< Dimension >(), i );
   }

protected:
   // setters for boundary tags
   template< int Dimension >
   void boundaryTagsSetEntitiesCount( const GlobalIndexType& entitiesCount )
   {
      BaseType::setEntitiesCount( DimensionTag< Dimension >(), entitiesCount );
   }
      
   template< int Dimension >
   void resetBoundaryTags()
   {
      BaseType::resetBoundaryTags( DimensionTag< Dimension >() );
   }

   template< int Dimension >
   void updateBoundaryIndices()
   {
      BaseType::updateBoundaryIndices( DimensionTag< Dimension >() );
   }

   template< int Dimension >
   __cuda_callable__
   void setIsBoundaryEntity( const GlobalIndexType& entityIndex, bool isBoundary )
   {
      static_assert( WeakTrait< Dimension >::boundaryTagsEnabled, "You try to access boundary tags which are not configured for storage." );
      BaseType::setIsBoundaryEntity( DimensionTag< Dimension >(), entityIndex, isBoundary );
   }
};

} // namespace BoundaryTags
} // namespace Meshes
} // namespace TNL
